/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: ftfile.c,v 1.6 2001/02/20 04:43:52 maf Exp $
 */

#if HAVE_CONFIG_H
 #include <config.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include <dirent.h>
#include <sys/stat.h>
#include <syslog.h>
#include <limits.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STRINGS_H
 #include <strings.h>
#endif
#if HAVE_STRING_H
  #include <string.h>
#endif
#include <time.h>
#include <fcntl.h>
#include "ftlib.h"
#include "report.h"

/*
 * function: ftfile_scandir
 * 
 * Load directory contents into the file ager data structures
 * Files that do not match the flow-tools naming convention or
 * do not have the correct magic word in the header are not loaded
 * to prevent accidental removal.
 *
 * returns: < 0 error
 *          >= 0 ok
*/
int ftfile_scandir(struct ftfile_ager *fa, char *dir)
{
  DIR *dirp;
  struct dirent *dirent;
  int fd, ret, done;
  struct stat sb;
  struct ftfile_entry *n1, *n2;
  extern int errno;
  extern int debug;
  struct fts2header head;

  ret = -1;

  TAILQ_INIT(&fa->head);

  if (chdir (dir) == -1) {
    report(LOG_ERR, "chdir(%s): %s", dir, strerror(errno));
    return -1;
  }

  if (!(dirp = opendir("."))) {
    report(LOG_ERR, "opendir() .");
    return -1;
  }

  for (dirent = readdir(dirp); dirent; dirent = readdir(dirp)) {

    /* skip . and .. */
    if (dirent->d_name[0] == '.')
      if (!dirent->d_name[1])
        continue;
      if (dirent->d_name[1] == '.')
        if (!dirent->d_name[2])
          continue;

    /* skip anything that doesn't begin with "ft" "cf" and "tmp" */
    if ((strncmp(dirent->d_name, "ft", 2)) &&
       (strncmp(dirent->d_name, "cf", 2)) &&
       (strncmp(dirent->d_name, "tmp", 3))) {
      report(LOG_INFO, "ignoring: %s", dirent->d_name);
      continue;
    }

    /* make sure the file is actually a flow file */
    if ((fd = open (dirent->d_name,  O_RDONLY, 0)) == -1) {
      report(LOG_INFO, "open(%s): %s", dirent->d_name, strerror(errno));
      continue;
    }

    if (ftheader_read(fd, &head) < 0) {
      report(LOG_INFO, "ftheader_read(): failed.");
      close(fd);
      continue;
    }

#if BYTE_ORDER == BIG_ENDIAN
    if (head.byte_order == FT_HEADER_LITTLE_ENDIAN) {
      fts2header_swap(&head);
    }
#endif /* BYTE_ORDER == BIG_ENDIAN */
  
#if BYTE_ORDER == LITTLE_ENDIAN
    if (head.byte_order == FT_HEADER_BIG_ENDIAN) {
      fts2header_swap(&head);
    }
#endif /* BYTE_ORDER == LITTLE_ENDIAN */

    if (fstat(fd, &sb) == -1) {
      report(LOG_INFO, "stat(%s): %s", dirent->d_name, strerror(errno));
      close (fd);
      continue;
    }

    close (fd);

    /* can only store proper length files below */
    if (strlen(dirent->d_name) > FT_FILELEN) {
      report(LOG_INFO, "bad filename: %s", dirent->d_name);
      continue;
    }

    /* insert the entry in the list sorted by start time of flow file */
    done = 0;
    TAILQ_FOREACH(n1, &fa->head, chain) {

      if (n1->start > head.start) {

        if (!(n2 = malloc(sizeof (struct ftfile_entry)))) {
          report(LOG_ERR, "malloc(): failed");
          goto errout;
        }

        n2->size = sb.st_size;
        n2->start = head.start; 
        strcpy((char*)&n2->name, (char*)dirent->d_name);

        TAILQ_INSERT_BEFORE(n1, n2, chain);
        done = 1;
        break;
      }

    } /* TAILQ_FOREACH */

    if (!done) {
      if (!(n2 = malloc(sizeof (struct ftfile_entry)))) {
        report(LOG_ERR, "malloc(): failed");
        goto errout;
      }
      n2->size = sb.st_size;
      n2->start = head.start; 
      strcpy((char*)&n2->name, (char*)dirent->d_name);
      TAILQ_INSERT_TAIL(&fa->head, n2, chain);
    }

    fa->num_bytes += sb.st_size;
    fa->num_files ++;

  }  /* for dirent */

  ret = 0;

errout:
  closedir(dirp);

  if (debug)
    report(LOG_INFO, "ftfile_scandir(): loaded %lu files", fa->num_files);

  return ret;

} /* ftfile_scandir */

/*
 * function: ftfile_add_tail
 *
 * Add a file to the end of the list
 *
 * returns: < 0 error
 *          >= 0 ok
*/
int ftfile_add_tail(struct ftfile_ager *fa, char *fname, off_t size,
  u_int32 start)
{
  struct ftfile_entry *n1;

  if (!(n1 = malloc(sizeof (struct ftfile_entry)))) {
    report(LOG_ERR, "malloc(): failed");
    return -1;
  }
  n1->size = size;
  n1->start = start;
  strcpy((char*)&n1->name, fname);
  TAILQ_INSERT_TAIL(&fa->head, n1, chain);
  fa->num_files ++;
  fa->num_bytes += size;

  return 0;

} /* ftfile_add_tail */

/*
 * function: ftfile_expire
 *
 * If doit is set, and the directory has exceeded the maximum size
 * of files, or maximum storage size remove files until the limits
 * are under the mark.  curbytes is a hint (fudge factor) to account
 * for the existing open flow export.
 *
 * returns: < 0 error
 *          >= 0 ok
*/
int ftfile_expire (struct ftfile_ager *fa, int doit, int curbytes)
{
  u_int i;
  struct ftfile_entry *n1;
  u_int64 bytes;
  extern int errno, debug;

  /*
   * if max_files is set, remove files starting at the head of the list until
   * max_files <= num_files.  update num_files, num_bytes
   */

  i = 0;
  bytes = 0;

  if (fa->max_files && (fa->num_files > fa->max_files)) {
    TAILQ_FOREACH(n1, &fa->head, chain) {
      report(LOG_INFO, "remove/1 %s", n1->name);
      bytes += n1->size;
      ++i;
      if (doit) {
        TAILQ_REMOVE(&fa->head, n1, chain);
        free(n1);
        if (unlink((char*)n1->name) == -1) 
          report(LOG_ERR, "unlink(%s): %s", n1->name, strerror(errno));
      } /* doit */
      if ((fa->num_files - i) <= fa->max_files)
        break;
    } /* TAILQ_FOREACH */
    if (doit) {
      fa->num_files -= i;
      fa->num_bytes -= bytes;
    } /* doit */
  } /* if */

  if (debug)
    report(LOG_INFO, "remove/1 %u files", i);

  i = 0;
  bytes = 0;

  /*
   * if max_bytes is set, remove files starting at the head of the list until
   * max_bytes <= num_bytes
   */

  if (fa->max_bytes && (fa->num_bytes+curbytes > fa->max_bytes)) {
    TAILQ_FOREACH(n1, &fa->head, chain) {
      report(LOG_INFO, "remove/2 %s", n1->name);
      bytes += n1->size;
      ++i;
      if (doit) {
        TAILQ_REMOVE(&fa->head, n1, chain);
        free(n1);
        if (unlink((char*)n1->name) == -1) 
          report(LOG_ERR, "unlink(%s): %s", n1->name, strerror(errno));
      } /* doit */
      if ((fa->num_bytes+curbytes - bytes) <= fa->max_bytes)
        break;
    } /* TAILQ_FOREACH */
    if (doit) {
      fa->num_files -= i;
      fa->num_bytes -= bytes;
    } /* doit */
  } /* if */

  if (debug)
    report(LOG_INFO, "remove/2 %u files", i);

  return 0;

} /* ftfile_expire */

/*
 * function: ftfile_dump
 *
 * Dump the contents of the file ager data struct.
 *
 * returns: < 0 error
 *          >= 0 ok
*/
int ftfile_dump(struct ftfile_ager *fa)
{
  struct ftfile_entry *n1;

  TAILQ_FOREACH(n1, &fa->head, chain) {

    report(LOG_INFO, "name=%s  size=%ld  time=%lu", n1->name, (long)n1->size,
      (unsigned long)n1->start);

  } /* TAILQ_FOREACH */

  return 0;

} /* ftfile_dump */
